	title "map aer from silicon retina to 4 orientation fields"
	title "S.C.Liu and T. Delbruck 10/99"
	title "10/14/99 10:44"
	title " added filter command with thomas"
	title " added stimulate command with shihchii and jorg"
	title " added RF project command with shihchii and thomas"

; case sensitivity here is ON by default

	processor 16C74
	include p16C74.inc

	__config _CP_OFF & _HS_OSC & _PWRTE_OFF & _WDT_ON


;***** VARIABLE DEFINITIONS
; context vars for interrupts
w_temp        EQU     0x20        ; variable used for context saving 
status_temp   EQU     0x21        ; variable used for context saving

; define registers for sc
; port A are request/ack
Handshake	equ	PORTA	; all AER handshake on porta
SReq		equ	0	; sender request to us
SAck		equ	1	; our ack to sender
RReq		equ	2 	; our req to reciever
RAck		equ	3	; receiver ack to us
PolBit		equ	5	; sex of sender pixel (on or off)
; porte will be spike output
Spike		equ	0	; on porte

;portb is sender
SPort		equ	PORTB	; sender address bits on portb

RPortX		equ	PORTC	; we use first 5 bits (32 addresses)
RPortY		equ	PORTD	; we use first 5 bits

SAddr		equ	0x22	; orig sender address
SAddrX		equ	0x23	; X part
SAddrY		equ	0x24	; Y part (shifted so it is in low order) 

RAddrX		EQU 	0x25	; register for x-address sender
RAddrY		EQU 	0x26	; register for y_address sender

Polarity	equ	0x27	; this holds 1 or 0 to indicate sender pixel sex

TableLength	equ	0x28	; length of projection table
lastchar	equ	0x29	; last char received on serial port

ProgMode	equ	0x2a	; program mode 0=project, 1=filter, 2=stimulate, 3=RF project
FiltAddr	equ	0x2b	; address to filter for
FiltPol		equ	0x2c	; sex of filter
StimX		equ	0x2d	; stimulation x address
StimY		equ	0x2e	; stimulation y address 
StimIntLow	equ	0x2f	; stimulation interval, low order byte
StimIntHigh	equ	0x30	; stimulation interval, high order byte
count		equ	0x31	; count for table
NumRFs		equ	0x32	; number of receptive fields stored in table, each is 6 bytes
TablePtr	equ	0x33	; pointer in RF table, used to march along table
TableStart	equ	0x34	; start of projection table

;;;;;;;;;;;;;;;; constants
LOADTABLECMD	equ	't'	; load table command, character tee
				; tnxyxyxy n=#projections, xyxy are relative addresses to project to
SETFILTERCMD	equ	'f'	; set filter command, character f
				; fab a=sender address to look for, b=sender polarity
STIMCMD		equ	's'	; stimulate command
				; sxylh xy are receiver addresses to stimulate, lw=2 byte, 16 bit interval in timer1 counts
RFPROJCMD	equ	'p'	; RF project command
				; p,n,m,  xl,yl,xh,yh,x,y, xl,yl,xh,yh,x,y
				; n is 6*number of RF's, m is number of RF's, xl,yl are lower corner of RF, xh,yh are upper corner. x,y is receiver address to stimulate, 
				; no commas/spaces in command.  each n, xl, x, etc. is one byte.
PROJMODE	equ	0
FILTMODE	equ	1
STIMMODE	equ	2
RFPROJMODE	equ	3

;serial port receive doesn't work robustly on C74 chip with BRGH (high speed) set
; max usable rate with 20MHz clock is 78600 at low baud rate setting,
; but linux doesn't support other than 19200, 38400, 56700 and 115200
; 19200 is highest rate on PIC with 20MHz clock that has acceptable rate error
SPBREGVAL19200	equ	d'15'	; for 19200 baud with 20MHz crystal with BRGH=0
SPBREGVAL9600	equ	d'32'	; for 9600 baud with 20MHz crystal with BRGH=0


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; macros
;bank switching macros
Bank0 macro ; go to bank 0
	bcf STATUS, RP0
;	bcf STATUS, RP1
	endm
	
Bank1 macro ; go to bank 1
	bsf STATUS, RP0
;	bcf STATUS, RP1  ;RP1 not needed for only bank1 on C74, always clear
	endm



;;;;;;;;;;;;;;;;;;;;;; code starts here (perhaps with interrupt service routine)

	ORG     0x000             ; processor reset vector
	clrf    PCLATH            ; ensure page bits are cleared
  	goto    main              ; go to beginning of program


;;;;;;;;;;;;;;;;;;;;interrupt service
;only 1 interrupt source, read char on serial port from computer
; put char read in lastchar variable

; first save context
ISR
	ORG     0x004             ; interrupt vector location
	movwf   w_temp            ; save off current W register contents
	movf	STATUS,w          ; move status register into W register
	clrf	STATUS            ; ensure file register bank set to 0
	movwf	status_temp       ; save off contents of STATUS register
	Bank0	

; get char in RCREG
	movf	RCREG,W
	movwf	lastchar
	sublw	LOADTABLECMD
	btfss	STATUS,Z	; see if it is load table command
	goto	checkfiltcmd	; if not, check for filter command
	call	LoadTable	; if so, load table
	goto	returnfromint	; and return to processing
	
checkfiltcmd
	movf	lastchar,W
	sublw	SETFILTERCMD
	btfss	STATUS,Z	; see if it is filter command
	goto	checkstimcmd	; if not, check for stimulate command	
	call	SetFilter	; if so, set filtering
	goto	returnfromint	; and return to processing

checkstimcmd	
	movf	lastchar,W
	sublw	STIMCMD
	btfss	STATUS,Z	; see if it is filter command
	goto	checkrfprojcmd	; if not, just ignore
	call	SetStimulate	; if so, set stimulation parameters
	goto	returnfromint	; and return to processing

checkrfprojcmd	
	movf	lastchar,W
	sublw	RFPROJCMD
	btfss	STATUS,Z	; see if it is filter command
	goto	returnfromint	; if not, just ignore
	call	SetRFProj	; if so, set rf projection parameters

returnfromint
; now restore context before returning from interrupt
	clrf    STATUS            ; ensure file register bank set to 0
	movf    status_temp,w     ; retrieve copy of STATUS register
	movwf	STATUS            ; restore pre-isr STATUS register contents
	swapf   w_temp,f
	swapf   w_temp,w          ; restore pre-isr W register contents

	retfie			;return also sets GIE (global interrupt enable)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; start of main code

main
	btfss	STATUS,NOT_TO
	goto	SetupPorts	; if NOT_TO is clear, WDT timed out, skip WDT initialization

;setup WDT prescaler
;c74 ref page 63
	clrf	TMR0		; clear timer0
	clrwdt
	Bank1
	movlw	b'10001101'	; now assign prescaler to WDT, set for medium timeout, about a half second
	movwf	OPTION_REG
	Bank0

;setup  ports
SetupPorts
	Bank1
	movlw	b'101001'
	movwf	TRISA
	movlw	b'110'
	movwf	ADCON1
	movlw	b'11111111'
	movwf	TRISB
	movlw	b'11000000'
	movwf	TRISC		; note bits 6,7 of portc are Tx,rx for serial port
				; they must be tristated
	clrf	TRISD
	clrf	TRISE		; porte are outputs
	Bank0
	clrf	Handshake
	clrf	RPortX
	clrf	RPortY

; set up timer for stimulate command
SetupTimer
	Bank0
	movlw	b'00110001'	
	movwf	T1CON		; init timer1 for 1/8 prescale (1.6us@20MHz/count)

; init serial port settings
	Bank0
	bsf RCSTA, SPEN ; enable serial port pins RX/TX
	bsf RCSTA, CREN ; enable reception of characters
	Bank1 
	; don't transmit in this program
	;bsf 	TXSTA, TXEN	; enable transmission of chars, transmit by writing to TXREG
	;don't set BRGH, high rate receive doesn't work on C74 robustly
	movlw 	SPBREGVAL9600	; baud rate setting, see header section
	movwf	SPBRG 		; set baud rate
	;enable interrupt on received chars
	bsf	PIE1,RCIE	; enable interrupts on rec char
	Bank0
	bsf	INTCON,PEIE	; enable peripheral interrupts
	bsf	INTCON,GIE	; enable interrupts globally

;if WDT timeout occured, don't reload projection table with default
	btfss	STATUS,NOT_TO
	goto	WDTimeOut	; if NOT_TO is clear, WDT timed out, skip initialization

; init some variables
; load table with default values
	movlw	1
	movwf	TableLength	; table is 1 long
	clrf	TableStart
	clrf	TableStart+1	; load table with default, 1:1 sender:receiver
	clrf	ProgMode	; default is to simply project 1:1, no filtering

; come here if wdt timeout
WDTimeout

; toggle acks on both sender and rec channels to make sure
; chips are not in some funny state
	bsf	Handshake,SAck
	bsf	Handshake,RReq
	bcf	Handshake,SAck
	bcf	Handshake,RReq

;;;;;;;;;; main event loop
MainLoop
	clrwdt			; clear before every event

	movf	ProgMode,w
	sublw	STIMMODE
	btfss	STATUS,Z	; see if stimulation mode
	goto	WaitSReqHigh	; if not, wait for request

;stimulate only
Stimulate
	movf	StimX,w
	movwf	RAddrX
	movf	StimY,w
	movwf	RAddrY		;load address to send to receiver to RAddr
				;registers so Send subroutine can send it

	call Send		; write to ports, do receiver handshake
	movf	StimIntLow,w
	movwf	TMR1L
	movf	StimIntHigh,w
	movwf	TMR1H		; load timer1, it counts up
	bcf	PIR1,TMR1IF	; clear overflow interupt flag bit
StimWait
	clrwdt
	btfss	PIR1,TMR1IF
	goto	StimWait	; wait till it overflows
	goto	MainLoop	; send address again
;previous loop will only be terminated by new command sent over serial port
; we go back to MainLoop to reset WDT and also check if some other mode
; was entered.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; wait here for sender request 
; come here anytime not purely stimulating (ignoring sender) 
WaitSReqHigh
	btfss	Handshake,SReq	; wait till sender gives us a spike
	goto	WaitSReqHigh

; got sender request
; now save sender address in reciever registers as X and Y part
	movf	SPort,w
	movwf	SAddr

;save polarity bit
	movf	Handshake,w
	andlw	b'00100000'
	movwf	Polarity

;sender address/polarity saved, now Acknowledge sender
	bsf	Handshake,SAck

; see what we want to do
	movf	ProgMode,f	; mov ProgMode to itself to set zero status if it is zero
	btfss	STATUS,Z	; if ProgMode=0, projecting so skip to Project
	goto	Filter		; if filtering

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; projecting spikes
Project
	movf	SAddr,w		; get sender address again
	andlw	b'00001111'
	movwf	SAddrX		;save sender X

	movf	SAddr,w
	andlw	b'11110000'	;get high 4 bits
	bcf	STATUS,C	; clear carry bit
	movwf	SAddrY
	rrf	SAddrY,f
	rrf	SAddrY,f
	rrf	SAddrY,f
	rrf	SAddrY,f	; save sender address Y

; now generate new addresses
	movf	TableLength,w	;get length of table
	movwf	count		;length of table in count (1 count is X and Y, 2 bytes)
	movlw	TableStart
	movwf	FSR		;start of table in FSR

SendLoop
	movf	SAddrX,w	; X part of address
	movwf	RAddrX		; prepare to work on it
	movf	INDF,w		; get X offset
	addwf	RAddrX,f	; add offset
	
	incf	FSR,f		; increment table pointer to point to Y offset
	
	movf	SAddrY,w	; Y part of address
	movwf	RAddrY		; prepare to work on it
	movf	INDF,w		; get Y offset
	addwf	RAddrY,f	; add offset

	incf	FSR,f		; increment pointer to point to next X offset

	call	Send		; send offsetted address to reciever
	decfsz	count,f		; decrement count
	goto	SendLoop

	goto WaitSReqLow	; when done projecting, finish sender communication

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; come here to filter spikes
Filter
	movf	ProgMode,w
	sublw	FILTMODE
	btfss	STATUS,Z	; see if stimulation mode
	goto	RFProj		; if not, wait for request

	movf	FiltAddr,w
	subwf	SAddr,w	
	btfss	STATUS,Z	; see if filter address is same as sender address
	goto 	WaitSReqLow	; if not, just acknowledge

; now test polarity
	movf	Polarity,w
	xorwf	FiltPol,w	; xor with polarity bit, dont destroy filtpol
	btfss	STATUS,Z	
	goto 	WaitSReqLow	; if not, just ackknowledge

; they match
	bsf	PORTE,Spike
	bcf	PORTE,Spike	; toggle spike bit, bit 0 on porte

	goto	WaitSReqLow
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; come here to project to receiver addresses from sender RF rectangles
RFProj
; first we get sender address in SAddrX and SAddrY, 
; for each table entry, check if sender addresses lies within rectangle,
; if so, send a spike to receiver address for that rectangle
	movf	SAddr,w		; get sender address again
	andlw	b'00001111'
	movwf	SAddrX		;save sender X

	movf	SAddr,w
	andlw	b'11110000'	;get high 4 bits
	bcf	STATUS,C	; clear carry bit
	movwf	SAddrY
	rrf	SAddrY,f
	rrf	SAddrY,f
	rrf	SAddrY,f
	rrf	SAddrY,f	; save sender address Y

; get start of rf table
	movf	NumRFs,w	;get length of table in number of 6 byte entries
	movwf	count		;num RFs=length of table in (6 byte entries) into count
	movlw	TableStart
	movwf	FSR		;start of table in FSR
	movwf	TablePtr	; save in TablePtr as well, for subsequent RFs

RFCheckAddr
; now check addresses against present table entry to see if it is within bounds of rectangle
; test in xlow, ylow, xhigh, yhigh order
; subwf, subtracts w from f,  if w<=f, carry bit is SET
; first xlow
	movf	INDF,w		; get x low
	subwf	SAddrX,w	; subtract xlow from saddrx, leave result in W
	btfss	STATUS,C	; test carry bit,
	goto	NextRF		; if Carry bit clear, xlow was bigger than saddrx, so sender spike was not in RF

;ylow
	incf	FSR,f		; point to ylow
	movf	INDF,w		; get ylow in w
	subwf	SAddrY,w	; sub ylow from saddry, leave result in w
	btfss	STATUS,C	; test carry again
	goto	NextRF		; if Carry bit clear, ylow was bigger than saddry, so sender spike was not in RF

;xhigh
	movf	SAddrX,w	; get saddrx in w
	incf 	FSR,f		; point to xhigh
	subwf	INDF,w		; subtract saddrx from xhigh, leave in w
	btfss	STATUS,C	; test carry again
	goto	NextRF		; if Carry bit clear, addrx was > xhigh, so sender spike was in RF
	
;yhigh
	movf	SAddrY,w	; get saddry in w
	incf 	FSR,f		; point to yhigh
	subwf	INDF,w		; subtract saddry from yhigh, leave in w
	btfss	STATUS,C	; test carry again
	goto	NextRF		; if Carry bit clear, addry was > yhigh, so sender spike was in RF
	

; spike in RF here, now send a spike to receiver
	incf	FSR,f		; point to receiver address X
	movf	INDF,w		; get it into w
	movwf	RAddrX		; save it
	incf	FSR,f		; point to rcvr Y
	movf	INDF,w		; get it
	movwf	RAddrY		; save it
	call	Send		; send the spike

; come here when done with sending a spike or if sender spike not in present RF
; we don't know where FSR points now
; take 
NextRF
	decfsz	count,f		; decrement RF count, skip if zero
	goto	MoreRFs
	goto	WaitSReqLow	; done with RF's, finish sender handshake

; come here when there are more than 1 RF to check
MoreRFs
	movf	TablePtr,w	; get present table pointer
	addlw	6		; add 6 to point to next table
	movwf	TablePtr	; put in pointer
	movwf	FSR		; and put in FSR for lookup
	goto	RFCheckAddr	; to check next RF against present sender spike

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; finally, finish sender handshake
WaitSReqLow
	btfsc	Handshake,SReq	; see if sender has taken away request yet (almost certainly has by now)
	goto	WaitSReqLow
	
	bcf	Handshake,SAck	; take away our acknowledge to sender
	
	goto 	MainLoop	; to next spike from sender
	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;subroutines
Send
; send addresses stored in RAddrX and RAddrY to receiver
	; check if RAck is low
	; write addresses to ports RPortX, RPortY, from registers RAddrX and RAddrY
	; pull REQ to receiver high
	; wait for ACK from receiver to go high
	; lower REQ
WaitRAckLow
	btfsc	Handshake,RAck
	goto	WaitRAckLow

	movf	RAddrX,w
	movwf	RPortX
	movf	RAddrY,w
	movwf	RPortY
	bsf	Handshake,RReq
WaitRAckHigh
	btfss	Handshake,RAck
	goto	WaitRAckHigh
	bcf	Handshake,RReq
	return

;-------------------------------------------------
LoadTable
; loads table with offsets for addresses
; this table load is for projection list, each sender spike projects to itself plus the offsets in the table

	clrf	ProgMode	; now we're projecting
	call 	GetChar		; get next character
	movwf	TableLength	; this is length of table
				; each table entry is 2 bytes
	movwf	count		; put in count
	bcf	STATUS,C	; clear carry bit to prepare X2
	rlf	count,f		; x2
				; now count holds 2 x tablelength

	movlw	TableStart
	movwf	FSR		; put address of start of table in FSR
projloadloop
	clrwdt
	call	GetChar		; get table entry (first X offset, then Y offset, then X, then Y, etc)
	movwf	INDF		; put char where FSR points
	incf	FSR,f		; point to next table entry
	decfsz	count,f		; decrement count
	goto	projloadloop	; if not zero, do next char

	return			; done loading table

;-------------------------------------------------
SetFilter
; loads table with offsets for addresses
	movlw	FILTMODE
	movwf	ProgMode	; 1 for filtering

	call 	GetChar		; get next character
	movwf	FiltAddr	; this is address to filter for
	call	GetChar
	bcf	STATUS,C
	movwf	FiltPol		; polarity of spike
	rlf	FiltPol,f	
	rlf	FiltPol,f
	rlf	FiltPol,f
	rlf	FiltPol,f	
	rlf	FiltPol,f	; PUT in same bit as polarity bit on porta	
	return			; done loading table

;-----------------------------------------------------
SetStimulate
	movlw	STIMMODE
	movwf	ProgMode	; 2 for stimulating

	call 	GetChar	
	movwf	StimX
	call	GetChar
	movwf	StimY
	call	GetChar
	movwf	StimIntLow
	call	GetChar
	movwf	StimIntHigh	; get parameters
	return
;-------------------------------------------------
SetRFProj
; loads table with offsets for addresses
; come here if we get RFPROJCMD ('p')

	movlw	RFPROJMODE
	movwf	ProgMode	; set program mode now

	call 	GetChar		; get next character
	movwf	TableLength	; this is total length of table in bytes
				; each table entry is 6 bytes, so tablelength must be multiple of 6
	movwf	count		; count holds total length of table in bytes, e.g. =12 for two projections

	call 	GetChar		; get next character
	movwf	NumRFs		; this is number of RF's (each 6 bytes) e.g. =2 for two projections

	movlw	TableStart
	movwf	FSR		; put address of start of table in FSR
rfloadloop
	clrwdt
	call	GetChar		; get table entry (first X offset, then Y offset, then X, then Y, etc)
	movwf	INDF		; put char where FSR points
	incf	FSR,f		; point to next table entry
	decfsz	count,f		; decrement count
	goto	rfloadloop	; if not zero, do next char

	return			; done loading table

;----------------------------------------------------
GetChar
; get next char from serial port, put in W and lastchar
	btfss	PIR1,RCIF	; see if char received
	goto	GetChar
	movf	RCREG,W
	movwf	lastchar
	return


	end
